/*
 *  main.c
 *  Derived from q4linux
 *
 *  Converts QuickTime movie files to AVI format
 *
 *  Modified by Benoit Joossen (July 2005).
 *  Company : Aero Quartet
 *
 *  You may copy, modify and redistribute this source file under
 *  the terms of the GNU Public License
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "qtprivate.h"
#include "funcprotos.h"
//#include "util.c"
#include <Carbon/Carbon.h>
#include <QuickTime/QuickTime.h>
#include "glue.h"

//define debugChunkMapping
//#define debugAudioSync
#define OpenDML
#define outputProgress
//#define debugChunkMapping
// GrandMaxNumberOfVideoEntries and AudioEntries was increased by a factor of 10 after discovering that the Bob Dylan movie had a keyframe only every 12 seconds !
#define GrandMaxNumberOfVideoEntries 3000
#define GrandMaxNumberOfAudioEntries 45000
#define VideoBufferSize 6400000
#define ReadBufferSize 4*1024*1024
// 640kb should be enough for anybody
#define MaxBalanceVideo 0.0
#define MinBalanceAudio 0.0
// The 2 parameters above will adjust audio/video sync. 25 14.4
#define SampleAlertSize 1000000
#define SmallDiscontinuityThreshold 5.0

// Find label XXXZ for dependencies on movie format. TATA for dependencies on multifile
// Movie must be a RefMovie with one video track and one or several audio tracks.
// All the media can come from one or several avi files, but video medias are merged into the same track.
// Audio tracks are usually not merged, but must be compatible (format, bitrate, channels, ..)
// Editing points must be on video keyframes. Edits don't need to be in chronological order.
// First track is video, second track is audio
// Audio codec must be .mp3 (VBR MPEG Layer 3). All channels, frequencies and bitrates are supported in theory.

// Max 100 files referenced.


// Function prototypes

int quicktime_init(quicktime_t *file);
quicktime_t* quicktime_open(char *filename, int rd, int wr);
FSSpec FSSpecWithString(char *string);
FILE* openDataReference(char * filename, short indexRef);
int getInfo(quicktime_t *file, char *toConvert);
int quicktime_write_frame(quicktime_t *file, unsigned char *video_buffer, int64_t bytes, int track);
int quicktime_audio_bits(quicktime_t *file, int track);
int quicktime_write_audio(quicktime_t *file, unsigned char *audio_buffer, long bytes, int track);
int convertMovie(quicktime_t *file, char *toConvert);
quicktime_t* quicktime_open(char *filename, int rd, int wr);
int quicktime_close(quicktime_t *file);
void reportError(char * msg);
int fasterRead(int refIndex, off_t offset64a, int64_t bytes, unsigned char *buffer);
int findMediaSampleSize(Media audioMedia);

// End of prototypes

int main (int argc,char** argv) {
	CFStringRef controlPath, outPath, inPath;
	int result = 1;
	if (argc != 4) {
		printf("quicktimeToAvi\n\nRequires 3 arguments\n#1 path of .mov file to convert\n#2 path of .avi file to write\n#3 path of control file\n");
		goto bail;
	}
	EnterMovies();
// Reading and checking parameters
	// Read path of .mov file to convert
	inPath = CFStringCreateWithCString(NULL, argv[1], CFStringGetSystemEncoding());
	// Read path of .avi file to write
	outPath = CFStringCreateWithCString(NULL, argv[2], CFStringGetSystemEncoding());
	// Read path of control file
	controlPath = CFStringCreateWithCString(NULL, argv[3], CFStringGetSystemEncoding());
    if (!inPath) { printf("Could not get CFString for arg#1 \n"); goto bail; }
	if (!outPath) { printf("Could not get CFString for arg#2\n"); goto bail; }
	if (!controlPath) { printf("Could not get CFString for arg#3\n"); goto bail; }
// Converting file
	quicktime_t *new_file = quicktime_open(argv[2], 0, 1);
	getInfo(new_file, argv[1]);
	quicktime_init_riff(new_file); // Writing header
	convertMovie(new_file, argv[1]); 
	result = quicktime_close(new_file);
	DisposeHandle((Handle)sdh);
	bail:
	ExitMovies();
	return result;
}


int quicktime_init(quicktime_t *file)
{
	bzero(file, sizeof(quicktime_t));
	file->cpus = 1;
	file->color_model = 5; //BC_RGB888;
	file->use_presave = 1;
	return 0;
}


FSSpec FSSpecWithString(char *string) { // 19 Dec 2004
	CFStringRef s = CFStringCreateWithCString ( kCFAllocatorDefault, string, kCFStringEncodingASCII);
	CFURLRef url = CFURLCreateWithFileSystemPath(kCFAllocatorDefault, (CFStringRef)s, kCFURLPOSIXPathStyle, false);
    FSRef ref;
    FSSpec spec;
    OSErr err = noErr;
    CFURLGetFSRef(url, &ref);
    err = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, &spec, NULL);
    if (err != noErr) printf("mov2avi : Error %d occured when getting the FSSpec", err);
    CFRelease(url);
	CFRelease(s);
    return spec;
}


FILE* openDataReference(char * filename, short indexRef) { // Retrieving sample info : offset, size, filename
	// Returns open .avi file (referenced from .mov file passed as arg)
	FSSpec theFileSpec = FSSpecWithString(filename);
	short resRefNum;
	Movie theMovie;
	Boolean dataRefWasChanged;
	OSErr err = OpenMovieFile (&theFileSpec, &resRefNum, fsRdPerm);
	err = NewMovieFromFile (&theMovie, resRefNum, 0, NULL, newMovieActive, &dataRefWasChanged);
	OSType typ;
	long dataRefAttributes;
	FSAliasInfoBitmap valid;
	CFStringRef refPath;
	Handle dataRef;
	short dataRefIndex = indexRef;
	Media theMedia = GetTrackMedia(GetMovieIndTrackType (theMovie, 1, VisualMediaCharacteristic, movieTrackCharacteristic)); // 16 Dec 2006, video track is suposed to contain all data references!
	err = GetMediaDataRef(theMedia, dataRefIndex, &dataRef, &typ, &dataRefAttributes);
	if (err != noErr) {
		printf("mov2avi openDataReference : Error %d retrieving movie info : cannot get dataRef\n",err);
		return NULL;
	}
	OSStatus oss = FSCopyAliasInfo ((AliasHandle) dataRef, NULL, NULL, &refPath, &valid, NULL);
	if (oss != noErr) {
		printf("mov2avi openDataReference : Error %d copying alias info\n", (int) oss);
		return NULL;
	}
	DisposeHandle(dataRef);
	char buffer[255];
	Boolean okCString = CFStringGetCString (refPath, buffer, 255, kCFStringEncodingUTF8);//CFStringGetSystemEncoding());
	if (okCString) {
		FILE *myFile = fopen(buffer, "rb");
		return myFile;
	}
	return NULL;
}


int getInfo(quicktime_t *file, char *toConvert) {
	short resRefNum;
	FSSpec theFileSpec;
	Boolean dataRefWasChanged;
	Movie theMovie;
	long numTracks;
	Track t;
	Media m;
	long dataOffset, size, sampleDescriptionIndex;
	TimeValue sampleTime, durationPerSample;
	Fixed ww, hh;
	OSErr err = noErr;
	theFileSpec = FSSpecWithString(toConvert);
	err = OpenMovieFile (&theFileSpec, &resRefNum, fsRdPerm);
	err = NewMovieFromFile (&theMovie, resRefNum, 0, NULL, newMovieActive, &dataRefWasChanged);
	numTracks = GetMovieTrackCount(theMovie);
	// TATA if (numTracks != 2) goto bail;
	file->total_length = GetMovieDuration(theMovie);
	// Extracting video information
	t = GetMovieIndTrackType (theMovie, 1, VisualMediaCharacteristic, movieTrackCharacteristic);
	if (t == NULL) goto bail;
	GetTrackDimensions(t, &ww, &hh);
	file->vtracks[0].track->tkhd.track_width = ww/65536;
	file->vtracks[0].track->tkhd.track_height = hh/65536;
	m = GetTrackMedia(t);

	SampleDescriptionHandle sampleDescriptionH = (SampleDescriptionHandle)NewHandleClear(4);
	err = GetMediaSampleReference (m, &dataOffset, &size, TrackTimeToMediaTime(0, t), &sampleTime, &durationPerSample, sampleDescriptionH, &sampleDescriptionIndex, 1, NULL, NULL); // 16 Dec 2007
	file->moov.trak[0]->mdia.mdhd.time_scale = GetMediaTimeScale(m);
	file->moov.trak[0]->mdia.minf.stbl.stts.table[0].sample_duration = durationPerSample;
	file->moov.trak[0]->mdia.minf.stbl.stts.table[0].sample_count = GetMediaDuration(m)/durationPerSample;
	ImageDescription *imgDescr = (* sampleDescriptionH);
	file->moov.trak[0]->mdia.minf.stbl.stsd.table[0].width = imgDescr->width;
	file->moov.trak[0]->mdia.minf.stbl.stsd.table[0].height = imgDescr->height;
	
	long * fformat = (long *) &(file->moov.trak[0]->mdia.minf.stbl.stsd.table[0].format);
	//*fformat = (** sampleDescriptionH).dataFormat;
	int vfourcc = (** sampleDescriptionH).dataFormat;
	DisposeHandle((Handle)sampleDescriptionH);
	// Video codec translation
	switch (vfourcc) {
		case 'dmb1': vfourcc = 'MJPG'; break;
		case 'dvcp': vfourcc = 'dvsd'; break;
		//case 'dvhp': vfourcc = 'DVH1'; break; // not verified
		//never worked case 'mjpa': vfourcc = 'JPEG'; break;
	}
	*fformat = CFSwapInt32HostToBig(vfourcc); // 16 Dec 2007
	
	// Extracting audio information
	t = GetMovieIndTrackType (theMovie, 1, AudioMediaCharacteristic, movieTrackCharacteristic+movieTrackEnabledOnly); // 07 Mar 2009, added Track enabled to handle correctly multi-audio movies.
	if (t == NULL) goto bail;
	m = GetTrackMedia(t);
	sdh = (SoundDescriptionHandle) NewHandle(0);
	long cc = GetMediaSampleDescriptionCount(m);
	if (cc != 1) printf("WARNING ! Audio has %d sample descriptions\n", cc); 
	err = GetMediaSampleReference (m, &dataOffset, &size, 0, &sampleTime, &durationPerSample, sdh, &sampleDescriptionIndex, 0, NULL, NULL); // maxNumberOfSamples was 1, changed to 0 to make work with 8bit integer mono, 05 July 2007
	mediaSampleSize = findMediaSampleSize(m); // 14 Feb 2008
	if (mediaSampleSize == -1) mediaSampleSize = size;
	printf("mediaSampleSize = %d\n", mediaSampleSize);

	file->moov.trak[1]->mdia.mdhd.time_scale = GetMediaTimeScale(m);
	file->moov.trak[1]->mdia.minf.stbl.stts.table[0].sample_duration = 1; //durationPerSample; // (+12) XXXZ changed was 1
	file->moov.trak[1]->mdia.minf.stbl.stts.table[0].sample_count = GetMediaDuration(m); //GetMediaDuration(m)/durationPerSample; // XXXZ changed was GetMediaDuration(m)
	file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].sample_size = 1; //(** sdh).sampleSize; //16; // (+26) XXXZ changed was 1.
	//file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].channels = 2; // (+24) XXXZ
	//file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].sample_rate = 48000; //(+32) XXXZ
	fformat = (long *) &(file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].format);

	*fformat = (** sdh).dataFormat;
	// ((** sdh).sampleRate)/65536 also gives sample rate 
	// QuickTime 7 required !
	AudioStreamBasicDescription inputDescription;
	err = QTSoundDescriptionGetProperty (sdh, kQTPropertyClass_SoundDescription,
kQTSoundDescriptionPropertyID_AudioStreamBasicDescription,
sizeof (AudioStreamBasicDescription),
&inputDescription, NULL );
	file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].sample_rate = inputDescription.mSampleRate;
	file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].channels = inputDescription.mChannelsPerFrame;
	file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].bytes_per_packet = inputDescription.mBytesPerPacket;
	
/*	if (1) { // 17 Feb 2008, to make work alaw stereo and twos mono 16 bit cases.
		long b0 = *sdh;
		int mBytesPerPacket = *((int *) (b0+0x24));
		mBytesPerPacket = CFSwapInt32LittleToHost(mBytesPerPacket);
		mBytesPerPacket &= 0x0000ffff;
		if (mBytesPerPacket != inputDescription.mFramesPerPacket) file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].bytes_per_packet = mBytesPerPacket;
	}*/
	file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].samples_per_packet = inputDescription.mFramesPerPacket;
	
	file->atracks[0].channels = inputDescription.mChannelsPerFrame; // 26 Nov 2006, uncommented 05 July 2007
	quicktime_codec_t *myCodec = file->atracks[0].codec;
	myCodec->fourcc = (**sdh).dataFormat;
	int fourcc = myCodec->fourcc;
	
	if ((fourcc&0xffff0000) == 0x6d730000) {
		myCodec->wav_id = fourcc&0x0000ffff;
	} else { // Translate fourcc to Microsoft WAVE codes
		switch ((**sdh).dataFormat) {
			case '.mp3': myCodec->wav_id = 0x55; break; // MPEG-1 layer 3
			case '.mp2': myCodec->wav_id = 0x50; audioSyncSupported = false; break; // MPEG-1 layer 2, or MPEG Audio, 16 Dec 2007
			case 'raw ': myCodec->wav_id = 0x01; break; // unsigned, linear PCM. 16-bit data is stored in little endian format.
			//case 'fl32': myCodec->wav_id = 0x03; break; // 32-bit floating point PCM. never verified
			//case 'fl64': myCodec->wav_id = 0x03; break; // 64-bit floating point PCM. never verified
			case 'alaw': myCodec->wav_id = 0x06; break;	// A-law
			case 'ulaw': myCodec->wav_id = 0x07; break; // mu-law
			//case 'ac2 ' : myCodec->wav_id = 0x30; break; // never verified
			case 'ac-3': myCodec->wav_id = 0x2000; break;//isAudioCompressed = true; break; // AC-3, A52, dolby, 5:1
			//case ' Ac3': myCodec->wav_id = 0x92; break; // never verified
			//case 'Qclp': myCodec->wav_id = 0x0150; break; // Qualcomm PureVoice never verified
			// 'QDM2': myCodec->wav_id = 0x0450; break; // QDesign Music 2 never verified
			case 'mp4a': myCodec->wav_id = 0xff; break;//isAudioCompressed = true; mediaSampleSize = 1024; break; // AAC format 
			case 'ima4': myCodec->wav_id = 0x11; break; // IMA 4 AVDPCM never verified
			//case 'in24': myCodec->wav_id = 0x01; break; // 24-bit, big endian, linear PCM.
			//case 'in32': myCodec->wav_id = 0x01; break; // 32-bit, big endian, linear PCM.
			case 'twos': myCodec->wav_id = 0x01; reverse16 = true; break; // signed, linear PCM. 16-bit data is stored in big endian format.
			case 'sowt': myCodec->wav_id = 0x01; break; // signed, linear PCM. 16-bit data is stored in little endian format.
		}
	}
	switch (myCodec->wav_id) { // 11 July 2007, because msxx also must set isAudioCompressed and other settings.
		case 0xff: mediaSampleSize = 1024;
		case 0x50:
		case 0x2000:
		case 0x55: isAudioCompressed = true;
			break;
		case 0x11: isAudioCompressed = true;
			isSpecialADPCM = true;
			audioSyncSupported = false; // 29 Dec 2007
			break;
		case 0x02: isAudioCompressed = true; // 03 Jan 2008
			isSpecialADPCM = true;
			copySampleDescription = true;
			audioSyncSupported = false;
			break;
	}
	
	copySampleDescription = true; // 03 Feb 2008
	//myCodec->wav_id = 0x50;//01;
	//yCodec->fourcc = (**sdh).dataFormat;//".mp2";//raw ";
	printf("Audio fourcc %4.4s and WAVE 0x%x\n", &fourcc, myCodec->wav_id);
	
	sampleSize = (** sdh).sampleSize;
	sampleRate = inputDescription.mSampleRate;
	packetSize = inputDescription.mBytesPerPacket;
	
	if (myCodec->wav_id == 0x55) { // 23 Feb 2008
		file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].bytes_per_packet = 0;
		file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].samples_per_packet = 1;
	}
	
	if (myCodec->wav_id == 0x2000) { // 06 Mar 2009
		file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].bytes_per_packet = 0;
		file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].samples_per_packet = 1;
	}
	
	/* The following snippet gives a human readable information like 
		"MPEG Layer 3 stereo 48,000Hz"
	char *buffer = malloc(200);
	CFStringRef descr = buffer;
	err = QTSoundDescriptionGetProperty (sdh, kQTPropertyClass_Audio, kQTAudioPropertyID_SummaryString, 200, &descr, NULL);
	free(buffer);
	*/
	// The following snippet calculates bitrate from media data size
	//but fails in some cases, like 3 avi files merged together (reports only data size of first two!)
	if (myCodec->wav_id == 0x55) { // Bitrate calculated from mp3 codec settings
		long b0 = *sdh;
		int bitrate = *((int *) (b0+0x58)); // 0x58 is offset to bitrate.
		bitrate = CFSwapInt32LittleToHost(bitrate);
		printf("Bit rate %d\n", 8*bitrate);
		audioFactor = bitrate; // Bitrate in bytes/seconds */
		// Rounding, example 7998.3 to 8000, 26 Nov 2006, QuickTime can change media scale to 7999 from 8000 when assembling a refMovie.
		
		audioFactor = floor((audioFactor/25.0)+0.5)*25;
	}
	// 07 Mar 2009, Perian 1.1.3 seems to return 0 for mp3, so we calculate audioFactor with second method...
	if (audioFactor==0 || (myCodec->wav_id != 0x55)) {// 03 Feb 2008 (fourcc != '.mp3') {
		wide w;
		long mediaSize;
		GetTrackDataSize64(t, 0, GetTrackDuration(t), &w); // 26 Nov 2006, instead of GetMediaDataSize that doesn't work fine when multiple data refs.
		mediaSize = w.lo; //GetMediaDataSize(m, 0, GetMediaDuration(m));
		float seconds = GetTrackDuration(t)/(0.0+GetMovieTimeScale(theMovie)); // 08 July 2007
		//GetMediaDuration(m)/(0.0+GetMediaTimeScale(m)); 
		audioFactor = mediaSize/seconds;//(0.125*seconds); // Bitrate in bytes/seconds */
	}
	printf("Byte rate %d\n", audioFactor);
	audioLever = inputDescription.mChannelsPerFrame*(quicktime_audio_bits(file, file->moov.trak[1])/8.0)/audioFactor;//(audioFactor*0.25); // Bitrate in bits/seconds inverted
	bail :
	DisposeMovie(theMovie);
	err = CloseMovieFile(resRefNum);
	return err;
}

int quicktime_write_frame(quicktime_t *file, unsigned char *video_buffer, int64_t bytes, int track) {
	//int64_t offset = quicktime_position(file);
	int result = 0;
	quicktime_atom_t chunk_atom;
	quicktime_video_map_t *vtrack = &file->vtracks[track];
	quicktime_trak_t *trak = vtrack->track;

	quicktime_write_chunk_header(file, trak, &chunk_atom);
	int mySize = bytes;
	result = !quicktime_write_data(file, video_buffer, mySize);
	quicktime_write_chunk_footer(file, 
					trak,
					vtrack->current_chunk,
					&chunk_atom, 
					1);
	file->vtracks[track].current_position++;
	file->vtracks[track].current_chunk++;
	videoSamples++;
	return result;
}

int quicktime_audio_bits(quicktime_t *file, int track) {
	if (isAudioCompressed) return 16; 
//	return 16;//8;	// XXXZ
	else //return file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].sample_size; // changed
	return sampleSize; // XXXZ
	if(file->total_atracks)
		return file->atracks[track].track->mdia.minf.stbl.stsd.table[0].sample_size;
	return 0;
}

int quicktime_write_audio(quicktime_t *file, unsigned char *audio_buffer, long bytes, int track) {
	int result;
	int64_t samples;
	quicktime_atom_t chunk_atom;
	quicktime_audio_map_t *track_map = &file->atracks[track];
	quicktime_trak_t *trak = track_map->track;

/* write chunk for 1 track */
	audioSamplesAccumulator += bytes * 8 / (0.0+quicktime_audio_bits(file, track)*file->atracks[track].channels);
	//printf("Audio accumulator %f, total %d\n", audioSamplesAccumulator, audioSamples);
	samples = floor(audioSamplesAccumulator);
	audioSamplesAccumulator -= samples;
	quicktime_write_chunk_header(file, trak, &chunk_atom);
	if (reverse16) {
		int i; char c;
		for (i=0; i<bytes; i+=2) {
			c = audio_buffer[i];
			audio_buffer[i] = audio_buffer[i+1];
			audio_buffer[i+1] = c;
		}
	}
	result = !quicktime_write_data(file, audio_buffer, bytes);
	quicktime_write_chunk_footer(file, 
					trak,
					track_map->current_chunk,
					&chunk_atom, 
					samples);

/*	file->atracks[track].current_position += samples; */
	file->atracks[track].current_chunk++;

	audioSamples += samples;
	return result;
}

float findFPS(Media videoMedia) { // 01 Jan 2008
	Track videoTrack = GetMediaTrack(videoMedia);
	SampleReference64Ptr sampleRefsVideo = calloc(1, sizeof(SampleReference64Record) * GrandMaxNumberOfVideoEntries);
	long actualNumberOfVideoEntries;
	float fps;
	// Determine fps (frames per second) NTSC=29.97 PAL=25.0
	OSErr err = GetMediaSampleReferences64 (videoMedia, 300+TrackTimeToMediaTime(0, videoTrack), NULL, NULL, NULL, GrandMaxNumberOfVideoEntries, &actualNumberOfVideoEntries, sampleRefsVideo);
	if ((err != noErr) || (actualNumberOfVideoEntries < 1)) return -1.0;
	if (actualNumberOfVideoEntries<100) printf("Few video entries used to determine fps: %d", actualNumberOfVideoEntries);
	int k, duration = 0, number = 0;
	for (k=0; k<actualNumberOfVideoEntries; k++) {
		duration += sampleRefsVideo[k].numberOfSamples*sampleRefsVideo[k].durationPerSample;
		number += sampleRefsVideo[k].numberOfSamples;
	}
	fps = number/(0.0+duration)*GetMediaTimeScale(videoMedia);
	printf("Video found, %f fps\n", fps);
	return fps;
}

int findMediaSampleSize(Media audioMedia) { // 14 Feb 2008
	long actualNumberOfAudioEntries;
	SampleReference64Ptr sampleRefsAudio = calloc(1, sizeof(SampleReference64Record) * 20);
	OSErr err = GetMediaSampleReferences64 (audioMedia, 0, NULL, NULL, NULL, 20, &actualNumberOfAudioEntries, sampleRefsAudio);
	if (err != noErr) {
		return -1;
	}
	int i, max = 0;
	for (i=0; i<actualNumberOfAudioEntries; i++) {
		if (sampleRefsAudio[i].dataSize>max) max = sampleRefsAudio[i].dataSize;
	}
	return max/1.39;
}

// Limitations :
// movie containing 1 video track and 1 audio track
// video track must be track #0, audio is #1

int convertMovie(quicktime_t *file, char *toConvert) {
	// Initializing stuff
	int videoRefIndex, audioRefIndex, oldAudioRefIndex = 0, index = 1, fill, result, videoInARow = 0;
	int segmentCorrection, audioSamplesBase;
	short resRefNum;
	float deltaSync, fps = 25.0;
	off_t offset64;
	FSSpec theFileSpec;
	Boolean forceFirstMove = false, dataRefWasChanged, needSync = false;
	Movie theMovie;
	OSErr err = noErr;
	theFileSpec = FSSpecWithString(toConvert);
	FILE *myMovieFile; 
	while (myMovieFile = openDataReference(toConvert, index)) { // open avi files (referenced from a mov or directly avi)
		refFiles[index++] = myMovieFile;
		printf("%d\n", index);
	}
	if (index == 1) reportError("No valid data reference found\n"); // 16 Feb 2008
	err = OpenMovieFile (&theFileSpec, &resRefNum, fsRdPerm);
	err = NewMovieFromFile (&theMovie, resRefNum, 0, NULL, newMovieActive, &dataRefWasChanged);
	//SetMovieTimeScale(theMovie, 90000); // 15 Dec 2007
	Track videoTrack = GetMovieIndTrackType (theMovie, 1, VisualMediaCharacteristic, movieTrackCharacteristic);
	Track audioTrack = GetMovieIndTrackType (theMovie, 1, AudioMediaCharacteristic, movieTrackCharacteristic+movieTrackEnabledOnly); // 07 Mar 2009
	Media videoMedia = GetTrackMedia(videoTrack);
	Media audioMedia = GetTrackMedia(audioTrack);
	long actualNumberOfVideoEntries, actualNumberOfAudioEntries, maxNumberOfVideoEntries, maxNumberOfAudioEntries;
	off_t lastVideoWritten = 0, lastAudioWritten = 0;
	SampleReference64Ptr sampleRefsVideo = calloc(1, sizeof(SampleReference64Record) * GrandMaxNumberOfVideoEntries);
	SampleReference64Ptr sampleRefsAudio = calloc(1, sizeof(SampleReference64Record) * GrandMaxNumberOfAudioEntries);
	unsigned char * video_buffer = calloc(1, VideoBufferSize);
	TimeValue correction = 0, myOldTrackTime, myTrackTime = 0, myChunkDuration, myChunkDurationToNextKeyframe, mySampleNextKeyframe, mySampleAudio, mySampleTime, time = 0;
	// Determine fps (frames per second) NTSC=29.97 PAL=25.0
	err = GetMediaSampleReferences64 (videoMedia, TrackTimeToMediaTime(0, videoTrack), NULL, NULL, NULL, 1, &actualNumberOfVideoEntries, sampleRefsVideo); // 16 Dec 2007
	if ((err != noErr) || (actualNumberOfVideoEntries < 1)) return -1;
	fps = findFPS(videoMedia); // 01 Jan 2008
	if (fps <= 0.0) reportError("Could not determine fps\n");
	globalFPS = fps;
	//GetMediaTimeScale(videoMedia)/(0.0+sampleRefsVideo[0].durationPerSample)*sampleRefsVideo[0].numberOfSamples;
	//
	// Main Loop : From sync sample to sync sample.
	//
	while ((time>=0) && (time < GetTrackDuration(videoTrack))) { // We parse the video track
		unsigned i,j, jDuration;
		//MoviesTask(NULL, 0);
		#ifdef outputProgress
		fprintf(stdout, "Writing AVI <%5.2f>\n", time/(0.0+GetTrackDuration(videoTrack))*100);
		fflush(stdout);
		#endif
		myOldTrackTime = myTrackTime;
		GetTrackNextInterestingTime(videoTrack, ((time == 0) ? nextTimeEdgeOK : 0)+nextTimeSyncSample, time, FixRatio(1, 1), &myTrackTime, &myChunkDuration);// 23 Apr 2006
		#ifdef debugChunkMapping
		printf("\nCHUNK %d to %d. Balance %f to %d.\n", myTrackTime, myTrackTime+myChunkDuration, audioSamples*audioLever*fps, videoSamples);
		#endif
		if (myChunkDuration < (0.011*GetMovieTimeScale(theMovie))) { 
			fprintf(stdout, "Small chunk skipped\n");
			time++; continue; 
		} // 27 Nov 2006. To discard rounding errors in keyframe cutting. 1000 stands for much less than a frame (3600)
		// 15 Dec 2007, removed depedency in timescale = 90000
		// (0.3*GetMovieTimeScale(theMovie))
		// Getting video samples
		mySampleTime = TrackTimeToMediaTime(myTrackTime, videoTrack);
		maxNumberOfVideoEntries = myChunkDuration/(0.0+GetMovieTimeScale(theMovie))*fps+0.1; // YYY
		if (maxNumberOfVideoEntries > GrandMaxNumberOfVideoEntries) reportError("video entries overflow\n");
		GetMediaNextInterestingTime(videoMedia, nextTimeSyncSample, mySampleTime, FixRatio(1, 1), &mySampleNextKeyframe, &myChunkDurationToNextKeyframe);
		if (mySampleNextKeyframe != mySampleTime) {
		} else correction = 0;
		err = GetMediaSampleReferences64 (videoMedia, mySampleTime, &mySampleTime, NULL, &videoRefIndex, maxNumberOfVideoEntries, &actualNumberOfVideoEntries, sampleRefsVideo);
		if (err == noErr) { // Getting audio samples
			index = 1; // Determine which track provides the audio at myTrackTime.
			while (audioTrack = GetMovieIndTrackType (theMovie, index++, AudioMediaCharacteristic, movieTrackCharacteristic+movieTrackEnabledOnly)) { // 07 Mar 2009
				mySampleAudio = TrackTimeToMediaTime(myTrackTime, audioTrack);
				if (mySampleAudio != -1) {
					audioMedia = GetTrackMedia(audioTrack);
					break;
				}
			}
			if (audioTrack == NULL) { // 21 Jun 2006
				index = 1;
				printf("Trying to find the audio track at end of interval\n");
				while (audioTrack = GetMovieIndTrackType (theMovie, index++, AudioMediaCharacteristic, movieTrackCharacteristic+movieTrackEnabledOnly)) { // 07 Mar 2009
					mySampleAudio = TrackTimeToMediaTime(myTrackTime+myChunkDuration-1, audioTrack);
					if (mySampleAudio != -1) {
						audioMedia = GetTrackMedia(audioTrack);
						break;
					}
				}
				mySampleAudio -= myChunkDuration/(0.0+GetMovieTimeScale(theMovie))*GetMediaTimeScale(GetTrackMedia(audioTrack));
				lastAudioWritten = 0; // 28 Apr 2008
			}
			if (audioTrack != NULL) {
				//(audioTrack == NULL) && ((GetMovieDuration(theMovie) - myTrackTime)>GetMovieTimeScale(theMovie))) reportError("audio discontinuity\n");
				//mySampleAudio = TrackTimeToMediaTime(myTrackTime, audioTrack);
				TimeValue aSampleAudio = TrackTimeToMediaTime(myTrackTime+myChunkDuration, audioTrack);
				float dAudio = aSampleAudio - myChunkDuration/(0.0+GetMovieTimeScale(theMovie))*GetMediaTimeScale(GetTrackMedia(audioTrack)) - mySampleAudio;
				aSampleAudio = mySampleAudio;
				if (aSampleAudio<0) aSampleAudio = 0; // 30 Dec 2007
				maxNumberOfAudioEntries = (maxNumberOfVideoEntries+10)/fps*(audioFactor/sampleSize); //+0.1; // XXXZ
				if (maxNumberOfAudioEntries > GrandMaxNumberOfAudioEntries) { printf("audio entries overflow %d\n", maxNumberOfAudioEntries); reportError("audio entries overflow\n"); }
				
				err = GetMediaSampleReferences64 (audioMedia, aSampleAudio, &mySampleAudio, NULL, NULL, maxNumberOfAudioEntries, &actualNumberOfAudioEntries, sampleRefsAudio);
				//if (mySampleAudio != aSampleAudio) printf("#### %d vs %d\n", aSampleAudio, mySampleAudio);
				//mySampleAudio = aSampleAudio;
				audioRefIndex = videoRefIndex; // Video track dataRefs map perfectly the reference files. Thus, videoRefIndex tells us in which file the audio can be found.
				off_t offset64a = sampleRefsAudio[0].dataOffset.hi*65536*65536 + sampleRefsAudio[0].dataOffset.lo;
				if ((err == noErr) && (offset64a != 0)) {
					off_t audioGarbage = 0;
					i = 0; j = 0; jDuration = 0;
					if (oldAudioRefIndex == audioRefIndex) { // In case of file switch, we cannot look for continuity
						if (fabs(offset64a-lastAudioWritten)>10000000) {
							#ifdef debugAudioSync
							printf("Detected coarse address discontinuity (%lld to %lld) in audio at %d. Adjusting sync again\n", lastAudioWritten, offset64a, myTrackTime);
							#endif
							needSync = true;
						}
						segmentCorrection = 0;
						if (offset64a <= lastAudioWritten) { // Look for continuity. Avoid repeating already written samples
							for (i=0; i<maxNumberOfAudioEntries; i++) {
								offset64a = sampleRefsAudio[i].dataOffset.hi*65536*65536 + sampleRefsAudio[i].dataOffset.lo;
								if (offset64a == lastAudioWritten) {
									segmentCorrection += sampleRefsAudio[i].numberOfSamples;
									i++;
									break;
								}
							}
							if (i == maxNumberOfAudioEntries) {
								i = 0;
								#ifdef debugAudioSync
								printf("Detect lack of audio overlap at %d. Adjusting audio sync\n", myTrackTime);
								#endif
								needSync = true; //reportError("Audio overlap not found !\n");
							} else {//printf("Correction audio %d\n", ++i);
								//if (i>8) audioFactor -= 5000*i/(1+audioSamples);
								//if (i<4) audioFactor += 5000*(4-i)/(1+audioSamples);
								//audioLever = 1.0/(audioFactor*0.25);
								//printf("i == %d, correting audioFactor to %d\n", i, audioFactor);
							}
						}
					} else {
						#ifdef debugAudioSync
						printf("Detected change of refIndex %d to %d at %d, adjusting audio sync\n", oldAudioRefIndex, audioRefIndex, myTrackTime);
						#endif
						needSync = true;
					}
					if (needSync && audioSyncSupported) { // Change of audioRefIndex or media cut, so adjust audio sync
						// 16 Dec 2007. Added audioSyncSupported
						lastAudioWritten = 0;
						deltaSync = audioSamples*audioLever-videoSamples/(0.0+fps);
						int deltaBytes = ((aSampleAudio-mySampleAudio)/(0.0+GetMediaTimeScale(audioMedia))+deltaSync)*audioFactor;
						#ifdef debugAudioSync
						printf("Ajusting audio sync ... %d bytes\n", deltaBytes);
						#endif
						if (deltaBytes < 0) { // 25 Jul 2006. Negative sync
							audioGarbage = -deltaBytes;
							//audioOffset = deltaBytes; // We add some garbage audio to fill
							deltaBytes = 0;
						} else { // Positive sync
							//i = 0; // XXX just to test Currahee. Remove this later
							while ((sampleRefsAudio[i].dataSize<deltaBytes) && (i<=maxNumberOfAudioEntries)) {
								deltaBytes -= sampleRefsAudio[i].dataSize;
								i++;
							} // We remove samples and adjust the fraction of the first sample.
							if (i>maxNumberOfVideoEntries) {
								i = maxNumberOfVideoEntries; // Don't add any audio here.
							} else {
								sampleRefsAudio[i].dataSize -= deltaBytes;
								sampleRefsAudio[i].dataOffset.lo += deltaBytes;
								int lessSamples = deltaBytes/(0.0+sampleSize);
								sampleRefsAudio[i].numberOfSamples -= lessSamples;
								//audioOffset = deltaBytes; //forceFirstMove = true; videoInARow = 0;
								printf("Less samples ... %d\n", lessSamples);
							}
						}
						segmentCorrection = 0;
					}
					segmentCorrection = 0;
					//segmentCorrection += mySampleAudio/(0.0+GetMediaTimeScale(audioMedia))/audioLever;
					audioSamplesBase = 0;//audioSamples;
					needSync = false;
					oldAudioRefIndex = audioRefIndex;
					//
					// Secondary Loop : Samples interleaving
					//
					offset64 = sampleRefsVideo[0].dataOffset.hi*65536*65536 + sampleRefsVideo[0].dataOffset.lo;
					while (jDuration<actualNumberOfVideoEntries) { // We want exactly all the video frames 
						deltaSync = (audioSamples-audioSamplesBase+sampleRefsAudio[i].numberOfSamples+segmentCorrection)*audioLever*fps-videoSamples;
						//deltaSync = (audioSamples+sampleRefsAudio[i].numberOfSamples+segmentCorrection)*audioLever*fps-videoSamples;
						//if (audioOffset > 0) deltaSync = -0.01;
						if (videoInARow == 0) {
							if (((deltaSync <= MaxBalanceVideo) || forceFirstMove) && (i<actualNumberOfAudioEntries)) { // Need audio samples
								forceFirstMove = false;
								offset64a = sampleRefsAudio[i].dataOffset.hi*65536*65536 + sampleRefsAudio[i].dataOffset.lo;// + audioOffset;
								//audioOffset = 0;
								myMovieFile = refFiles[audioRefIndex];
								if (sampleRefsAudio[i].dataSize>SampleAlertSize) reportError("Audio sample size abnormally big\n");
								else {
									videoInARow = 0;
									if (audioGarbage>0) {
										printf("Audio Garbage %\n", audioGarbage);
										//result = fasterRead(audioRefIndex, offset64a, audioGarbage, video_buffer); // 13 Feb 2008
										//int k;
										//for (k=0; k<audioGarbage; k++) *(video_buffer+k) = 0x0;
										quicktime_write_audio(file, video_buffer, audioGarbage, 0);
										audioGarbage = 0;
									}
									int64_t bytes = sampleRefsAudio[i].dataSize;
									if (bytes > (1.4*mediaSampleSize)) bytes = mediaSampleSize;
									sampleRefsAudio[i].dataSize -= bytes;
									sampleRefsAudio[i].dataOffset.lo += bytes;
									#ifdef debugChunkMapping
									printf("..AUDIO dataOffset %lld dataSize %lld\n", offset64a, bytes);
									#endif
									result = fasterRead(audioRefIndex, offset64a, bytes, video_buffer);
									quicktime_write_audio(file, video_buffer, bytes, 0); 
									if (sampleRefsAudio[i].sampleFlags == 0) 
									quicktime_set_idx1_keyframe(file, file->atracks[0].track, audioFrames, audioFrames);
									else quicktime_noKeyframe_last_ix_record(file, file->atracks[0].track);
									audioFrames++;
									if (sampleRefsAudio[i].dataSize == 0) { 
										i++;
										lastAudioWritten = offset64a;
									}
								}
							} else videoInARow = 1;
						} // end if (videoInARow == 0);
						deltaSync = (audioSamples-audioSamplesBase+segmentCorrection)*audioLever*fps-videoSamples-sampleRefsVideo[j].numberOfSamples; // XXXZ
						//if (audioOffset > 0) deltaSync = -0.01;
						if (videoInARow > 0) {
							if ((deltaSync >= MinBalanceAudio) || (i==actualNumberOfAudioEntries)) {
								offset64 = sampleRefsVideo[j].dataOffset.hi*65536*65536 + sampleRefsVideo[j].dataOffset.lo;
								//printf("VIDEO dataOffset %lld dataSize %d duration %d\n", offset64, sampleRefsVideo[j].dataSize, sampleRefsVideo[j].durationPerSample);
								//if (sampleRefsVideo[j].dataSize == 0) j++;
								//else 
								if (sampleRefsVideo[j].dataSize>SampleAlertSize) reportError("Video sample size abnormally big\n");
								else {
									myMovieFile = refFiles[videoRefIndex];
									result = fasterRead(videoRefIndex, offset64, sampleRefsVideo[j].dataSize, video_buffer);
									//result = fseeko(myMovieFile, offset64, SEEK_SET);
									//result = fread(video_buffer, sampleRefsVideo[j].dataSize, 1, myMovieFile);
									int64_t bytes = sampleRefsVideo[j].dataSize;
									quicktime_write_frame(file, video_buffer, bytes, 0);
									lastVideoWritten = offset64+sampleRefsVideo[j].dataSize;
									if (sampleRefsVideo[j].sampleFlags == 0) {
										deltaSync = (audioSamples-audioSamplesBase+segmentCorrection)*audioLever*fps-videoSamples; // XXXZ
										//printf("\n\nKeyframe sync = %f\n\n", deltaSync);
										quicktime_insert_keyframe(file, videoSamples-1, 0);
										#ifdef debugChunkMapping
										printf("<< KEYFRAME >>\n");
										#endif
									} else quicktime_noKeyframe_last_ix_record(file, file->vtracks[0].track);
									#ifdef debugChunkMapping
										printf("VIDEO dataOffset %lld dataSize %d duration %d\n", offset64, sampleRefsVideo[j].dataSize, sampleRefsVideo[j].numberOfSamples);
									#endif
									int frames = (GetMediaTimeScale(videoMedia)/(fps-0.05))/sampleRefsVideo[j].durationPerSample; // 10 Apr 2007
									if (frames != 1) { // 06 Apr 2007
										// When a sample has duration > 1, probably it contains several 0 bytes frames
										// This is a trick used to put 25fps in a 30fps time base.
										// jDuration controls accumulated duration from 0 to actualVideoSamples,
										// j is the index for table query.
										unsigned kkk;
										for (kkk = 1; kkk<frames; kkk++) {
											quicktime_write_frame(file, video_buffer, 0, 0);
											quicktime_noKeyframe_last_ix_record(file, file->vtracks[0].track);
											jDuration++;
										}
									}
									videoInARow++; j++; jDuration++;
									
								}							
							} else {
								if (videoInARow == 1) forceFirstMove = true; // Solve interlock
								videoInARow = 0;
							}
						} // end videoInARow > 0
					} // end secondary loop
					//printf("Audio %d (%d)\n",actualNumberOfAudioEntries-i, actualNumberOfAudioEntries);
				} else { // end ((err == noErr) && (offset64a != 0))
					printf("Error %d getting audio samples\n", err);
				}
				if (fabs(dAudio)>SmallDiscontinuityThreshold) {
					printf("Detected a small audio discontinuity\n");
					needSync = true;
				}
			} else {// end audioTrack != NULL
				printf("Audio discontinuity\n");
				needSync = true;
			}
		} // end no error, get audio sample
		if ((myTrackTime + myChunkDuration - 2) > (time+1)) time = myTrackTime + myChunkDuration - 2; else time++;
	} // end main loop while ()
	return 0;
}

void reportError(char * msg) {
	printf("FATAL ERROR\n");
	printf(msg);
	exit(1);
}

quicktime_t* quicktime_open(char *filename, int rd, int wr) {
	quicktime_t *new_file = calloc(1, sizeof(quicktime_t));
	//char flags[10];
	int result = 0;
	quicktime_init(new_file);
	new_file->wr = 1;
	new_file->rd = 0;
	new_file->mdat.atom.start = 0;
	new_file->use_avi = 1;
	new_file->moov.total_tracks = 2;
	new_file->total_vtracks = 1;
	new_file->total_atracks = 1;
	new_file->moov.trak[0] = calloc(1, sizeof(quicktime_trak_t));
	new_file->moov.trak[1] = calloc(1, sizeof(quicktime_trak_t));
	new_file->atracks = (quicktime_audio_map_t*)calloc(1, sizeof(quicktime_audio_map_t));
	new_file->vtracks = (quicktime_video_map_t*)calloc(1, sizeof(quicktime_video_map_t));
	new_file->vtracks[0].track = new_file->moov.trak[0];
	new_file->atracks[0].track = new_file->moov.trak[1];
	//new_file->vtracks[0].track->tkhd.track_width = 320;
	//new_file->vtracks[0].track->tkhd.track_height = 240;
	new_file->moov.trak[0]->mdia.minf.is_video = 1;
	new_file->moov.trak[1]->mdia.minf.is_video = 0;
	//new_file->moov.trak[0]->mdia.mdhd.time_scale = 90000;
	new_file->moov.trak[0]->mdia.minf.stbl.stts.table = calloc(1, sizeof(quicktime_stss_table_t));
	//new_file->moov.trak[0]->mdia.minf.stbl.stts.table[0].sample_duration = 100;
	new_file->moov.trak[0]->mdia.minf.is_audio = 0;
	new_file->moov.trak[1]->mdia.minf.is_audio = 1;
	new_file->moov.trak[0]->mdia.minf.stbl.stsd.table = calloc(1, sizeof(quicktime_stsd_table_t));
	//new_file->moov.trak[0]->mdia.minf.stbl.stsd.table[1].sample_size = 100;
	//new_file->total_length = 100000;
	new_file->moov.trak[0]->tkhd.track_id = 1;
	new_file->moov.trak[1]->tkhd.track_id = 2;
	//new_file->moov.trak[0]->mdia.minf.stbl.stsd.table[0].format[0] = 65;
	new_file->atracks[0].codec = calloc(1, sizeof(quicktime_codec_t));
	/* 05 July 2007 new_file->atracks[0].channels = 1; // XXXZ
	//quicktime_codec_t *myCodec = new_file->atracks[0].codec;
	//myCodec->wav_id = 0x55; // Stands for MP3 XXXZ
	//myCodec->fourcc = ".mp3"; // XXXZ */
	new_file->moov.trak[1]->mdia.minf.stbl.stts.table = calloc(1, sizeof(quicktime_stss_table_t));
	new_file->moov.trak[1]->mdia.minf.stbl.stsd.table = calloc(1, sizeof(quicktime_stsd_table_t));
	//new_file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].channels = 2;
	//new_file->moov.trak[1]->mdia.minf.stbl.stsd.table[0].sample_rate = 44100;
	new_file->moov.trak[0]->mdia.minf.stbl.stss.table = calloc(1, sizeof(quicktime_stts_table_t) * 100);
	
	result = quicktime_file_open(new_file, filename, rd, wr);
	return new_file;
}

int quicktime_close(quicktime_t *file)
{
	int result = 0;
	if(file->wr) {
		if(file->use_avi) {
			//quicktime_atom_t junk_atom;
			//int i;
			//int64_t position = quicktime_position(file);
			#ifdef debugChunkMapping2
			int pos = position;
			printf("End position %d\n", pos);
			#endif
		// Finalize last header
			quicktime_finalize_riff(file, file->riff[file->total_riffs - 1]);
			//position = quicktime_position(file);
		#ifdef OpenDML
		// Finalize the odml header
			quicktime_finalize_odml(file, &file->riff[0]->hdrl);
		// Finalize super indexes
			quicktime_finalize_indx(file);
		// Pad ending
		/*
			quicktime_set_position(file, position);
			quicktime_atom_write_header(file, &junk_atom, "JUNK");
			for(i = 0; i < 0x406; i++)
				quicktime_write_int32_le(file, 0);
			quicktime_atom_write_footer(file, &junk_atom);
		*/
		#endif
		}

	}

	quicktime_file_close(file);

	//quicktime_delete(file);
	free(file);
	printf("Operation complete!\n");
	return result;
}

int fasterRead(int refIndex, off_t offset64a, int64_t bytes, unsigned char *buffer) {
/* Alternative: direct read
	FILE *myMovieFile = refFiles[refIndex];
	int result = fseeko(myMovieFile, offset64a, SEEK_SET);
	result = fread(buffer, bytes, 1, myMovieFile);
	return result; */
	int result;
	if (refIndex <= mediaRefs) {
		if (bufferPos[refIndex]>offset64a) { // Missed cache, we read directly
			FILE *myMovieFile = refFiles[refIndex];
			int result = fseeko(myMovieFile, offset64a, SEEK_SET);
			result = fread(buffer, bytes, 1, myMovieFile);
			return result;
		}
	} else while (refIndex > mediaRefs) { // Allocate buffer. Modified to while 16 Dec 2006
		readBuffer[++mediaRefs] = calloc(1, ReadBufferSize);
		bufferPos[mediaRefs] = -1; // Mark as uninitialized
	}
	if ((bufferPos[refIndex]>offset64a) || ((bufferPos[refIndex]+ReadBufferSize-offset64a-bytes)<0) || (bufferPos[refIndex]<0)) { // Need to read chunk
		//printf("XXXBuffering\n");
		FILE *myMovieFile = refFiles[refIndex];
		off_t pos = offset64a;
		if (pos>=(ReadBufferSize/2)) pos-=(ReadBufferSize/2);
		result = fseeko(myMovieFile, pos, SEEK_SET);
		result = fread(readBuffer[refIndex], ReadBufferSize, 1, myMovieFile);
		bufferPos[refIndex] = pos;
	}
	memcpy(buffer, readBuffer[refIndex]+offset64a-bufferPos[refIndex], bytes);
	return result;
}

int fasterWrite() {

}